# $Id: distributor_execprogram.py,v 1.2 2006/03/29 12:38:33 hat Exp $
#
import select,sys
from general import spawn, Connection

class Distributor_exec_program:
    """Wrapper class of Distributor to hide pipe to Exec_program object"""

    def __init__(self,execprog,maxdelay=None):
	(pid,r,w)=spawn()
	if pid==0:
	    #print "Distributor_exec_program.__init__: child"
	    # child process
	    self._execprog=execprog
	    self._maxdelay=maxdelay
	    self._connection=Connection(r,w)
	    try:
		self._child_run()
	    except KeyboardInterrupt:
		pass
	    sys.exit(0)
	else:
	    #print "Distributor_exec_program.__init__: parent"
	    # parent process. No need for maxdelay.
	    self._execprog=execprog	# needed for the hostname
	    self._connection=Connection(r,w)
	    self._results=[]
	    self._job=None  # last given job (for crash-recovery)
	    self._have_job=0	# I have a job for the child
	    self._child_busy=0	# child is idle
	    self._closedown=0	# we should close down

    # Communication over the pipe:
    # Parent(distributor)     Child(exec_program)
    #                     <-- ('close',)	Request to close down
    #	       ('close',) -->			Command to close down
    #       ('job',<job>) -->			Command to compute <job>
    #                     <-- ('res',<result>)	Result of computation
    #

    # child stuff (runs the exec_program object)

    def _child_run(self):
	self._execprog.open()	# run the open method of exec_program
	_have_result=0 # we have no result waiting
	while 1:
	    #print "Child loopie"
	    # wait for something to happen
	    inl=[self._connection.rhandle()]
	    if _have_result:
		outl=[self._connection.whandle()]
	    else:
		outl=[]
	    
	    if self._maxdelay!=None:
		(inl,outl,errl)=select.select(inl,outl,[], self._maxdelay)
	    else:
		(inl,outl,errl)=select.select(inl,outl,[])

	    if len(inl)==0 and len(outl)==0: # timeout
		self._connection.send(('close',))   # send request to be closed
	    if len(inl)>0:
		#print "Waiting for something from distributor"
		j=self._connection.recv()
		#print "forwarder_execprogram: received", repr(j)
		if j==None:
		    break	# connection to parent died, so quit
		if j[0]=='close':   # parent wants us to die
		    break
		if j[0]=='job':
		    assert _have_result==0
		    the_job=j[1]
		    result=the_job
		    ret_val=self._execprog.compute(the_job.job_type(),
					     the_job.parameters())
		    result.set_result(ret_val)
		    # result available in 'result'
		    _have_result=1
	    if len(outl)>0:
		assert _have_result==1
		self._connection.send(('res',result))
		_have_result=0
	self._execprog.close()
	self._connection.close()

    # Parent stuff (at the distributor side)

    def collect_fds(self):
	if self._have_job or self._closedown:	# we have something to send
	    #print "Distributor_exec_program.collect_fds(): add outgoing"
	    outl=[self._connection.whandle()]	# to the child
	else:
	    #print "Distributor_exec_program.collect_fds(): NO outgoing"
	    outl=[]
	#print "collect=",repr(([self._connection.whandle()],outl,None))
	return ([self._connection.rhandle()],outl,None)

    def handle_fds(self,inl,outl):
	"""Handle IO. return 0 if nothing interesting happened, 1 for quit"""
	#print "handle=",repr((inl,outl))
	if self._connection.rhandle() in inl:
	    j=self._connection.recv()
	    if j==None:	# child crashed
		self._connection.close()
		return 1
	    if j[0]=='res':
		self._results.append(j[1])
		self._child_busy=0
		return 0	# store result for query later
	    if j[0]=='close':	# child wants to die
		# honour the request if child is idle and we have no results
		if self._child_busy==0 and len(self._results)==0 and \
		                                          self._have_job==0:
		    self._connection.send( ('close',) ) # honour the request
		    self._connection.close()
		    return 1    # and die too
		return 0    # child is busy -> not dy please !
	    assert 0	# should never happen
	#print "outl available", self._connection.whandle() in outl
	if self._have_job and self._connection.whandle() in outl:
	    #print "Sending job to child_exec", repr(self._job)
	    self._connection.send(('job',self._job))
	    self._have_job=0
	    self._child_busy=1
	    return 0
	if self._closedown and self._connection.whandle() in outl:
	    self._connection.send( ('close',) )
	    self._connection.close()
	    return 1    # and die too
	return 0

    #
    # Actions that the Distributor object can do to a exec_prog object.
    #

    def close(self):
	"""Distributor requested shutdown of (idle) exec"""
	self._closedown=1

    def get_result(self):
	n=len(self._results)
	if n==0:
	    return None	# no result
	if n==1:
	    r=self._results[0]
	    self._results=[]
	    return r
	assert 0    # should never happen

    def put_job(self,j):
	#print "Distributor_exec_program.put_job("+repr(j)+")"
	assert not self._child_busy
	self._job=j
	self._have_job=1

    def read_job(self):
	"""Return the last provided job (so we still have a copy in case the
	   child crashed).
	   [Currently not used afaik]"""
	return self._job

    def get_hostname(self):
	"""Return the hostname of the exec_program object"""
	return self._execprog.get_hostname()

    def can_compute(self,jobtype):
	"""Return whether this exec_program can compute jobs of type
	   'jobtype'"""
	return self._execprog.can_compute(jobtype)

